/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Page management
*  SCOPE :              Page management
*  FILE :               page.hxx
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma interface
#endif
#ifndef __page_HXX
#define __page_HXX

/* IMPORTS *******************************************************************/

#ifndef __class_HXX
#include "class.hxx"
#endif

#ifndef __yoofunc_HXX
#include "yoofunc.hxx"
#endif

#ifndef __yootypes_H
#include "yootypes.h"
#endif

#ifndef __define_H
#include "define.h"
#endif

#ifndef __config_H
#include "config.h"
#endif

#ifndef __oid_HXX
#include "oid.hxx"
#endif

#ifndef __obj_HXX
#include "obj.hxx"
#endif

#ifndef __sockrpc_HXX
#include "sockrpc.hxx"
#endif

/*****************************************************************************/

class FreeBloc : public Bloc {  
  public :
  
  short& next() ;
} ;


struct Slot {
  char data[SLOT_SIZE] ;
  FreeBloc* slot() ;
} ;


class PageGen {

  protected :

  int _usedSpace ;     // Count used space in the page. 0 means free page
  int _pgAfter ;       // Number of pages that constitute a long object
  OID _cluster ;       // References cluter of the page
  PID _pid ;           // Store the absolute address of the page 
  short _firstFreeSlot ; // slot of the first free block in the free list
  int _error ;         // Return error for communication
#if defined(__alpha)  || defined( _vmpage_8192)
  u_short _tabObj[MAX_DIR_PG] ; // array of indirection
#else
  u_char _tabObj[MAX_DIR_PG] ; // array of indirection
#endif

  PageGen() : _usedSpace(0), _pgAfter(0) {}

  public :

  void setError(int err) { _error = err ; }
  int getError() { return _error ; } 
} ;

class Page : public PageGen {

  private :
  
  Slot _tab[SLOT_IN_PG] ; // Data in the page. Could be free slot or object
  unsigned short allocTab() ; // find a free indice in the indirection table
  unsigned short firstFree(); // find the first free slot in the page
  u_short firstUsed(u_short) ; // find the first used block from fp slot
  void move(u_short, u_short); // Move an object from op to np

  public :

  void init(Cluster*, PID pid, int npage = 1) ; // Initialise a new page
  int size() ;               // Number of pages if it is a long object
  int freeSpace() ;          // Return total freespace in the page
  PID& pid() ;               // PID of the page
  ObjectX* object(int);      // return the nth object of the page
  ObjectX* allocObject(int); // Allocation in the page
  void freeObject(OID obj);  // delete the object obj 
  void receive() ;           // compact the page and initialise objects
  Cluster* getCluster() ;    // Return cluster of the page
  int isempty() ;            // Return TRUE if it is an empty page
  int consistency(PID pid = NULL_PID,int* no=NULL,int* ndo=NULL,int* sz=NULL,
		  int* objPerClass=NULL) ;
} ;


inline short& FreeBloc::next() {
  return _next ;
}

inline FreeBloc* Slot::slot() {
  return (FreeBloc*)this ;
}

inline ObjectX* Page::object(int num) {
  return this->_tabObj[num] < SLOT_IN_PG ?
    (ObjectX*)&(this->_tab[this->_tabObj[num]]) : NULL ;
} 

inline PID& Page::pid() {
  return this->_pid ;
}

inline int Page::size() {
  return this->_pgAfter + 1 ;
}

inline int Page::freeSpace() { 
  return SLOT_IN_PG - this->_usedSpace ; 
}

inline SvcSocket& operator<<(SvcSocket& s, Page& p) { 
  s.send((char*)&p, PAGE_SIZE) ; return s ;
}

inline SvcSocket& operator>>(SvcSocket& s, Page& p) { 
  s.recv((char*)&p, PAGE_SIZE) ; return s ;
}

inline int Page::isempty() {
  return (_usedSpace == 0) && (_pgAfter == 0) ;
}

#endif
